# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: fipa.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\nfipa.proto\x12\x17\x61\x65\x61.fetchai.fipa.v1_0_0"\xc5\x0c\n\x0b\x46ipaMessage\x12J\n\x06\x61\x63\x63\x65pt\x18\x05 \x01(\x0b\x32\x38.aea.fetchai.fipa.v1_0_0.FipaMessage.Accept_PerformativeH\x00\x12\\\n\x0f\x61\x63\x63\x65pt_w_inform\x18\x06 \x01(\x0b\x32\x41.aea.fetchai.fipa.v1_0_0.FipaMessage.Accept_W_Inform_PerformativeH\x00\x12\x44\n\x03\x63\x66p\x18\x07 \x01(\x0b\x32\x35.aea.fetchai.fipa.v1_0_0.FipaMessage.Cfp_PerformativeH\x00\x12L\n\x07\x64\x65\x63line\x18\x08 \x01(\x0b\x32\x39.aea.fetchai.fipa.v1_0_0.FipaMessage.Decline_PerformativeH\x00\x12\x44\n\x03\x65nd\x18\t \x01(\x0b\x32\x35.aea.fetchai.fipa.v1_0_0.FipaMessage.End_PerformativeH\x00\x12J\n\x06inform\x18\n \x01(\x0b\x32\x38.aea.fetchai.fipa.v1_0_0.FipaMessage.Inform_PerformativeH\x00\x12V\n\x0cmatch_accept\x18\x0b \x01(\x0b\x32>.aea.fetchai.fipa.v1_0_0.FipaMessage.Match_Accept_PerformativeH\x00\x12h\n\x15match_accept_w_inform\x18\x0c \x01(\x0b\x32G.aea.fetchai.fipa.v1_0_0.FipaMessage.Match_Accept_W_Inform_PerformativeH\x00\x12L\n\x07propose\x18\r \x01(\x0b\x32\x39.aea.fetchai.fipa.v1_0_0.FipaMessage.Propose_PerformativeH\x00\x1a(\n\x0b\x44\x65scription\x12\x19\n\x11\x64\x65scription_bytes\x18\x01 \x01(\x0c\x1a\x1c\n\x05Query\x12\x13\n\x0bquery_bytes\x18\x01 \x01(\x0c\x1aM\n\x10\x43\x66p_Performative\x12\x39\n\x05query\x18\x01 \x01(\x0b\x32*.aea.fetchai.fipa.v1_0_0.FipaMessage.Query\x1aZ\n\x14Propose_Performative\x12\x42\n\x08proposal\x18\x01 \x01(\x0b\x32\x30.aea.fetchai.fipa.v1_0_0.FipaMessage.Description\x1a\xa6\x01\n\x1c\x41\x63\x63\x65pt_W_Inform_Performative\x12Y\n\x04info\x18\x01 \x03(\x0b\x32K.aea.fetchai.fipa.v1_0_0.FipaMessage.Accept_W_Inform_Performative.InfoEntry\x1a+\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xb2\x01\n"Match_Accept_W_Inform_Performative\x12_\n\x04info\x18\x01 \x03(\x0b\x32Q.aea.fetchai.fipa.v1_0_0.FipaMessage.Match_Accept_W_Inform_Performative.InfoEntry\x1a+\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x94\x01\n\x13Inform_Performative\x12P\n\x04info\x18\x01 \x03(\x0b\x32\x42.aea.fetchai.fipa.v1_0_0.FipaMessage.Inform_Performative.InfoEntry\x1a+\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x15\n\x13\x41\x63\x63\x65pt_Performative\x1a\x16\n\x14\x44\x65\x63line_Performative\x1a\x1b\n\x19Match_Accept_Performative\x1a\x12\n\x10\x45nd_PerformativeB\x0e\n\x0cperformativeb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "fipa_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY._options = None
    _FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY._serialized_options = b"8\001"
    _FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY._options = None
    _FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY._serialized_options = (
        b"8\001"
    )
    _FIPAMESSAGE_INFORM_PERFORMATIVE_INFOENTRY._options = None
    _FIPAMESSAGE_INFORM_PERFORMATIVE_INFOENTRY._serialized_options = b"8\001"
    _globals["_FIPAMESSAGE"]._serialized_start = 40
    _globals["_FIPAMESSAGE"]._serialized_end = 1645
    _globals["_FIPAMESSAGE_DESCRIPTION"]._serialized_start = 791
    _globals["_FIPAMESSAGE_DESCRIPTION"]._serialized_end = 831
    _globals["_FIPAMESSAGE_QUERY"]._serialized_start = 833
    _globals["_FIPAMESSAGE_QUERY"]._serialized_end = 861
    _globals["_FIPAMESSAGE_CFP_PERFORMATIVE"]._serialized_start = 863
    _globals["_FIPAMESSAGE_CFP_PERFORMATIVE"]._serialized_end = 940
    _globals["_FIPAMESSAGE_PROPOSE_PERFORMATIVE"]._serialized_start = 942
    _globals["_FIPAMESSAGE_PROPOSE_PERFORMATIVE"]._serialized_end = 1032
    _globals["_FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE"]._serialized_start = 1035
    _globals["_FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE"]._serialized_end = 1201
    _globals[
        "_FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY"
    ]._serialized_start = 1158
    _globals[
        "_FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY"
    ]._serialized_end = 1201
    _globals["_FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE"]._serialized_start = 1204
    _globals["_FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE"]._serialized_end = 1382
    _globals[
        "_FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY"
    ]._serialized_start = 1158
    _globals[
        "_FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY"
    ]._serialized_end = 1201
    _globals["_FIPAMESSAGE_INFORM_PERFORMATIVE"]._serialized_start = 1385
    _globals["_FIPAMESSAGE_INFORM_PERFORMATIVE"]._serialized_end = 1533
    _globals["_FIPAMESSAGE_INFORM_PERFORMATIVE_INFOENTRY"]._serialized_start = 1158
    _globals["_FIPAMESSAGE_INFORM_PERFORMATIVE_INFOENTRY"]._serialized_end = 1201
    _globals["_FIPAMESSAGE_ACCEPT_PERFORMATIVE"]._serialized_start = 1535
    _globals["_FIPAMESSAGE_ACCEPT_PERFORMATIVE"]._serialized_end = 1556
    _globals["_FIPAMESSAGE_DECLINE_PERFORMATIVE"]._serialized_start = 1558
    _globals["_FIPAMESSAGE_DECLINE_PERFORMATIVE"]._serialized_end = 1580
    _globals["_FIPAMESSAGE_MATCH_ACCEPT_PERFORMATIVE"]._serialized_start = 1582
    _globals["_FIPAMESSAGE_MATCH_ACCEPT_PERFORMATIVE"]._serialized_end = 1609
    _globals["_FIPAMESSAGE_END_PERFORMATIVE"]._serialized_start = 1611
    _globals["_FIPAMESSAGE_END_PERFORMATIVE"]._serialized_end = 1629
# @@protoc_insertion_point(module_scope)
