# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: models.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cmodels.proto\x12\x19\x61\x65\x61.helpers.search.models\"\xc1\x19\n\x05Query\x1a\xc0\x01\n\tAttribute\x12\x0c\n\x04name\x18\x01 \x01(\t\x12=\n\x04type\x18\x02 \x01(\x0e\x32/.aea.helpers.search.models.Query.Attribute.Type\x12\x10\n\x08required\x18\x03 \x01(\x08\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\"?\n\x04Type\x12\n\n\x06\x44OUBLE\x10\x00\x12\x07\n\x03INT\x10\x01\x12\x08\n\x04\x42OOL\x10\x02\x12\n\n\x06STRING\x10\x03\x12\x0c\n\x08LOCATION\x10\x04\x1an\n\tDataModel\x12\x0c\n\x04name\x18\x01 \x01(\t\x12>\n\nattributes\x18\x02 \x03(\x0b\x32*.aea.helpers.search.models.Query.Attribute\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x1a$\n\x08Location\x12\x0b\n\x03lon\x18\x01 \x01(\x01\x12\x0b\n\x03lat\x18\x02 \x01(\x01\x1a\x99\x01\n\x05Value\x12\x10\n\x06string\x18\x01 \x01(\tH\x00\x12\x10\n\x06\x64ouble\x18\x02 \x01(\x01H\x00\x12\x11\n\x07\x62oolean\x18\x03 \x01(\x08H\x00\x12\x11\n\x07integer\x18\x04 \x01(\x03H\x00\x12=\n\x08location\x18\x05 \x01(\x0b\x32).aea.helpers.search.models.Query.LocationH\x00\x42\x07\n\x05value\x1aN\n\x08KeyValue\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x35\n\x05value\x18\x02 \x01(\x0b\x32&.aea.helpers.search.models.Query.Value\x1a\x80\x01\n\x08Instance\x12\x39\n\x05model\x18\x01 \x01(\x0b\x32*.aea.helpers.search.models.Query.DataModel\x12\x39\n\x06values\x18\x02 \x03(\x0b\x32).aea.helpers.search.models.Query.KeyValue\x1a+\n\nStringPair\x12\r\n\x05\x66irst\x18\x01 \x01(\t\x12\x0e\n\x06second\x18\x02 \x01(\t\x1a(\n\x07IntPair\x12\r\n\x05\x66irst\x18\x01 \x01(\x03\x12\x0e\n\x06second\x18\x02 \x01(\x03\x1a+\n\nDoublePair\x12\r\n\x05\x66irst\x18\x01 \x01(\x01\x12\x0e\n\x06second\x18\x02 \x01(\x01\x1a\x83\x01\n\x0cLocationPair\x12\x38\n\x05\x66irst\x18\x01 \x01(\x0b\x32).aea.helpers.search.models.Query.Location\x12\x39\n\x06second\x18\x02 \x01(\x0b\x32).aea.helpers.search.models.Query.Location\x1a\xa1\x02\n\x05Range\x12\x42\n\x0bstring_pair\x18\x01 \x01(\x0b\x32+.aea.helpers.search.models.Query.StringPairH\x00\x12@\n\x0cinteger_pair\x18\x02 \x01(\x0b\x32(.aea.helpers.search.models.Query.IntPairH\x00\x12\x42\n\x0b\x64ouble_pair\x18\x03 \x01(\x0b\x32+.aea.helpers.search.models.Query.DoublePairH\x00\x12\x46\n\rlocation_pair\x18\x04 \x01(\x0b\x32-.aea.helpers.search.models.Query.LocationPairH\x00\x42\x06\n\x04pair\x1aW\n\x08\x44istance\x12\x39\n\x06\x63\x65nter\x18\x01 \x01(\x0b\x32).aea.helpers.search.models.Query.Location\x12\x10\n\x08\x64istance\x18\x02 \x01(\x01\x1a\xca\x01\n\x08Relation\x12\x44\n\x08operator\x18\x01 \x01(\x0e\x32\x32.aea.helpers.search.models.Query.Relation.Operator\x12\x35\n\x05value\x18\x02 \x01(\x0b\x32&.aea.helpers.search.models.Query.Value\"A\n\x08Operator\x12\x06\n\x02\x45Q\x10\x00\x12\x06\n\x02LT\x10\x01\x12\x08\n\x04LTEQ\x10\x02\x12\x06\n\x02GT\x10\x03\x12\x08\n\x04GTEQ\x10\x04\x12\t\n\x05NOTEQ\x10\x05\x1a\xca\x05\n\x03Set\x12?\n\x08operator\x18\x01 \x01(\x0e\x32-.aea.helpers.search.models.Query.Set.Operator\x12;\n\x06values\x18\x02 \x01(\x0b\x32+.aea.helpers.search.models.Query.Set.Values\x1a\xa5\x04\n\x06Values\x12\x45\n\x06string\x18\x01 \x01(\x0b\x32\x33.aea.helpers.search.models.Query.Set.Values.StringsH\x00\x12\x45\n\x06\x64ouble\x18\x02 \x01(\x0b\x32\x33.aea.helpers.search.models.Query.Set.Values.DoublesH\x00\x12\x44\n\x07\x62oolean\x18\x03 \x01(\x0b\x32\x31.aea.helpers.search.models.Query.Set.Values.BoolsH\x00\x12\x43\n\x07integer\x18\x04 \x01(\x0b\x32\x30.aea.helpers.search.models.Query.Set.Values.IntsH\x00\x12I\n\x08location\x18\x05 \x01(\x0b\x32\x35.aea.helpers.search.models.Query.Set.Values.LocationsH\x00\x1a\x16\n\x04Ints\x12\x0e\n\x06values\x18\x01 \x03(\x03\x1a\x19\n\x07\x44oubles\x12\x0e\n\x06values\x18\x01 \x03(\x01\x1a\x19\n\x07Strings\x12\x0e\n\x06values\x18\x01 \x03(\t\x1a\x17\n\x05\x42ools\x12\x0e\n\x06values\x18\x01 \x03(\x08\x1a\x46\n\tLocations\x12\x39\n\x06values\x18\x01 \x03(\x0b\x32).aea.helpers.search.models.Query.LocationB\x08\n\x06values\"\x1d\n\x08Operator\x12\x06\n\x02IN\x10\x00\x12\t\n\x05NOTIN\x10\x01\x1a\xc3\x06\n\x0e\x43onstraintExpr\x12\x41\n\x03or_\x18\x01 \x01(\x0b\x32\x32.aea.helpers.search.models.Query.ConstraintExpr.OrH\x00\x12\x43\n\x04\x61nd_\x18\x02 \x01(\x0b\x32\x33.aea.helpers.search.models.Query.ConstraintExpr.AndH\x00\x12\x43\n\x04not_\x18\x03 \x01(\x0b\x32\x33.aea.helpers.search.models.Query.ConstraintExpr.NotH\x00\x12P\n\nconstraint\x18\x04 \x01(\x0b\x32:.aea.helpers.search.models.Query.ConstraintExpr.ConstraintH\x00\x1aI\n\x02Or\x12\x43\n\nexpression\x18\x01 \x03(\x0b\x32/.aea.helpers.search.models.Query.ConstraintExpr\x1aJ\n\x03\x41nd\x12\x43\n\nexpression\x18\x01 \x03(\x0b\x32/.aea.helpers.search.models.Query.ConstraintExpr\x1aJ\n\x03Not\x12\x43\n\nexpression\x18\x01 \x01(\x0b\x32/.aea.helpers.search.models.Query.ConstraintExpr\x1a\xa0\x02\n\nConstraint\x12\x16\n\x0e\x61ttribute_name\x18\x01 \x01(\t\x12\x34\n\x04set_\x18\x02 \x01(\x0b\x32$.aea.helpers.search.models.Query.SetH\x00\x12\x38\n\x06range_\x18\x03 \x01(\x0b\x32&.aea.helpers.search.models.Query.RangeH\x00\x12=\n\x08relation\x18\x04 \x01(\x0b\x32).aea.helpers.search.models.Query.RelationH\x00\x12=\n\x08\x64istance\x18\x05 \x01(\x0b\x32).aea.helpers.search.models.Query.DistanceH\x00\x42\x0c\n\nconstraintB\x0c\n\nexpression\x1a\x88\x01\n\x05Model\x12\x44\n\x0b\x63onstraints\x18\x01 \x03(\x0b\x32/.aea.helpers.search.models.Query.ConstraintExpr\x12\x39\n\x05model\x18\x02 \x01(\x0b\x32*.aea.helpers.search.models.Query.DataModelB\x02H\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'models_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'H\001'
  _globals['_QUERY']._serialized_start=44
  _globals['_QUERY']._serialized_end=3309
  _globals['_QUERY_ATTRIBUTE']._serialized_start=54
  _globals['_QUERY_ATTRIBUTE']._serialized_end=246
  _globals['_QUERY_ATTRIBUTE_TYPE']._serialized_start=183
  _globals['_QUERY_ATTRIBUTE_TYPE']._serialized_end=246
  _globals['_QUERY_DATAMODEL']._serialized_start=248
  _globals['_QUERY_DATAMODEL']._serialized_end=358
  _globals['_QUERY_LOCATION']._serialized_start=360
  _globals['_QUERY_LOCATION']._serialized_end=396
  _globals['_QUERY_VALUE']._serialized_start=399
  _globals['_QUERY_VALUE']._serialized_end=552
  _globals['_QUERY_KEYVALUE']._serialized_start=554
  _globals['_QUERY_KEYVALUE']._serialized_end=632
  _globals['_QUERY_INSTANCE']._serialized_start=635
  _globals['_QUERY_INSTANCE']._serialized_end=763
  _globals['_QUERY_STRINGPAIR']._serialized_start=765
  _globals['_QUERY_STRINGPAIR']._serialized_end=808
  _globals['_QUERY_INTPAIR']._serialized_start=810
  _globals['_QUERY_INTPAIR']._serialized_end=850
  _globals['_QUERY_DOUBLEPAIR']._serialized_start=852
  _globals['_QUERY_DOUBLEPAIR']._serialized_end=895
  _globals['_QUERY_LOCATIONPAIR']._serialized_start=898
  _globals['_QUERY_LOCATIONPAIR']._serialized_end=1029
  _globals['_QUERY_RANGE']._serialized_start=1032
  _globals['_QUERY_RANGE']._serialized_end=1321
  _globals['_QUERY_DISTANCE']._serialized_start=1323
  _globals['_QUERY_DISTANCE']._serialized_end=1410
  _globals['_QUERY_RELATION']._serialized_start=1413
  _globals['_QUERY_RELATION']._serialized_end=1615
  _globals['_QUERY_RELATION_OPERATOR']._serialized_start=1550
  _globals['_QUERY_RELATION_OPERATOR']._serialized_end=1615
  _globals['_QUERY_SET']._serialized_start=1618
  _globals['_QUERY_SET']._serialized_end=2332
  _globals['_QUERY_SET_VALUES']._serialized_start=1752
  _globals['_QUERY_SET_VALUES']._serialized_end=2301
  _globals['_QUERY_SET_VALUES_INTS']._serialized_start=2118
  _globals['_QUERY_SET_VALUES_INTS']._serialized_end=2140
  _globals['_QUERY_SET_VALUES_DOUBLES']._serialized_start=2142
  _globals['_QUERY_SET_VALUES_DOUBLES']._serialized_end=2167
  _globals['_QUERY_SET_VALUES_STRINGS']._serialized_start=2169
  _globals['_QUERY_SET_VALUES_STRINGS']._serialized_end=2194
  _globals['_QUERY_SET_VALUES_BOOLS']._serialized_start=2196
  _globals['_QUERY_SET_VALUES_BOOLS']._serialized_end=2219
  _globals['_QUERY_SET_VALUES_LOCATIONS']._serialized_start=2221
  _globals['_QUERY_SET_VALUES_LOCATIONS']._serialized_end=2291
  _globals['_QUERY_SET_OPERATOR']._serialized_start=2303
  _globals['_QUERY_SET_OPERATOR']._serialized_end=2332
  _globals['_QUERY_CONSTRAINTEXPR']._serialized_start=2335
  _globals['_QUERY_CONSTRAINTEXPR']._serialized_end=3170
  _globals['_QUERY_CONSTRAINTEXPR_OR']._serialized_start=2640
  _globals['_QUERY_CONSTRAINTEXPR_OR']._serialized_end=2713
  _globals['_QUERY_CONSTRAINTEXPR_AND']._serialized_start=2715
  _globals['_QUERY_CONSTRAINTEXPR_AND']._serialized_end=2789
  _globals['_QUERY_CONSTRAINTEXPR_NOT']._serialized_start=2791
  _globals['_QUERY_CONSTRAINTEXPR_NOT']._serialized_end=2865
  _globals['_QUERY_CONSTRAINTEXPR_CONSTRAINT']._serialized_start=2868
  _globals['_QUERY_CONSTRAINTEXPR_CONSTRAINT']._serialized_end=3156
  _globals['_QUERY_MODEL']._serialized_start=3173
  _globals['_QUERY_MODEL']._serialized_end=3309
# @@protoc_insertion_point(module_scope)
